% general program for learning optimal control (PIRL) as described in section 19.2
clear all
PLANT='singlewell';
PLANT='acrobot';
TIMEDEPENDENT=1;			% time dependent problem 
REWARD='endcost';
INITIALIZATION='random';		% initialization of x0 
NEIGHBORHOOD='size';		% number: set by number of points, size: set by size
n_traj=50;
n_ref=100;
dt=0.05;
nplot=min(floor(0.1*n_traj),10);
eta=1;

rand('state',2);
randn('state',2);


switch PLANT
case 'acrobot',
	n=4;					% dimension x
	m=1;					% dimension u, xi
	x0=[-pi/2,-pi/2,0,0]';
	dx0=0.0*ones(n,1);			% variance in intialization 
	nu=20;						% diagonal covariance matrix <xi_i xi_j> = nu dt delta_{ij}
	R=1;						% diagonal u^2 cost R sum_i u_i^2
	rho=1000;					% initial neighborhood size when NEIGHBORHOOD='size';
	rhomax=1;					% maximal neighborhood size when NEIGHBORHOOD='size';
	n_neighbor=200;				% number of neighboring points when NEIGHBORHOOD='number';
	T=1;						% horizon time
	a=50;						% size of quadratic end of path cost relative to control cost
	noise_factor=0.9;
case 'singlewell'
	n=1;
	m=1;
	T=1;						% horizon time
	x0=2*ones(n,1);
	dx0=0*ones(n,1);			% variance in intialization 
	nu=1;						% diagonal covariance matrix <xi_i xi_j> = nu dt delta_{ij}
	R=1;						% diagonal u^2 cost R sum_i u_i^2
	rho=1000;						% initial neighborhood size when NEIGHBORHOOD='size';
	rhomax=1;					% maximal neighborhood size when NEIGHBORHOOD='size';
	n_neighbor=0.2*n_traj;				% number of neighboring points when NEIGHBORHOOD='number';
	a=5;						% size of quadratic end of path cost
	noise_factor=0.95;
end;

% end definitions -------------------------------------------------------
N=T/dt;
lambda=nu*R;

figure(20)
clf
x=zeros(n,N);				% stores a single trajectory
xall=zeros(n,N,n_traj);	% stores all trajectories
u=zeros(m,N);
Sall=zeros(N,n_traj);		% stores path costs of trajectories
phiall=zeros(1,n_traj);
ssall=zeros(N,n_ref);		% stores effective system size of local set of trajectories
dx1all=zeros(N,n_ref);	% stores size of neighborhood for each timeset and trajectory
lall=zeros(N,n_ref);	% stores number of states in neighborhood
height_s=zeros(n_traj,n_ref);	% stores end cost of stochastic trajectories
height_d=zeros(1,n_ref);		% stores end cost of deterministic trajectories
J=zeros(1,n_ref);		% stores expected cost to go
Jphi=zeros(1,n_ref);		% stores expected end cost
mean_increment=zeros(1,n_ref)	% stores av{xi} increment
std_increment=zeros(1,n_ref)	% stores av{xi} increment
mean_u=zeros(1,n_ref)
std_u=zeros(1,n_ref)

% initialize all trials 
switch INITIALIZATION
case 'fixed'
	x0all=x0*ones(1,n_traj);
case 'random'
	x0all=x0*ones(1,n_traj)+dx0*randn(1,n_traj);
end;

for i=1:n_ref
	% generate trajectories under uncontrolled dynamics 
	xiall=sqrt(nu*dt)*randn(m,N,n_traj);	% stores all noise realizations
	for	j=1:n_traj,
		for t=0:N-1,
			if t==0,
				x_cur=x0;
			else
				x_cur=x(:,t);
			end
			x(:,t+1)=plant(x_cur,u(:,t+1)*dt+xiall(:,t+1,j),dt,PLANT);
			dSkall(t+1,j)=dSk(u(:,t+1),xiall(:,t+1,j),R,dt);
		end;
		%[Sk, Sv]=S(x,u, xiall(:,:,j), a, R, dt,PLANT,REWARD);
		%Skall(j)=Sk;
		phiall(j)=phi(x(:,N),a,PLANT);
		Sall(N,j)=dSkall(N,j)+phiall(j);
		for t=N-1:-1:1,
			Sall(t,j)=dSkall(t,j)+Sall(t+1,j);
		end;
		height_s(j,i)=-phiall(j)/a;
		xall(:,:,j)=x;
	end;
	Smin=min(Sall(1,:));
	phimin=min(phiall);
	J(i)=Smin-lambda*log(1/n_traj*sum(exp(-(Sall(1,:)-Smin)/lambda)));
	Jphi(i)=phimin-lambda*log(1/n_traj*sum(exp(-(phiall-phimin)/lambda)));
	max(height_s)
	i
	% compute deterministic solution
	x=zeros(n,N);
	for t=0:N-1,
		xi_traj=squeeze(xiall(:,t+1,:))'+u(:,t+1)*dt;
		if t==0,
			x_cur=x0;
			x_traj=x0all;
		else
			x_cur=x(:,t);
			if n==1,
				x_traj=squeeze(xall(:,t,:))';
			else
				x_traj=squeeze(xall(:,t,:));
			end
		end
		% computes xi according to Eq. 239 as the weighted average of u^\mu dt + d\xi^\mu
		[xi,dx1(t+1),l(t+1),ss(t+1)]=compute_xi(x_cur, x_traj,xi_traj,Sall(t+1,:),t, TIMEDEPENDENT, NEIGHBORHOOD, rho, rhomax,n_neighbor,lambda,PLANT);
		increment(:,t+1)=xi/dt-u(:,t+1);
		u(:,t+1)=u(:,t+1)+eta*increment(:,t+1);
		x(:,t+1)=plant(x_cur,xi,dt,PLANT);
	end;
	mean_increment(i)=mean(increment,2);
	std_increment(i)=std(increment,[],2);
	ssall(:,i)=ss;
	dx1all(:,i)=dx1;
	lall(:,i)=l;
	%[Sk, Sv]=S(x, zeros(m,N), zeros(m,N), a, R, dt, PLANT,REWARD);
	height_d(i)=-phi(x(:,N),a,PLANT)/a;
	[wsort iwsort]=sort(Sall(1,:),'ascend');
	iwsort=iwsort(1:nplot);

	switch PLANT
	case 'acrobot',
    	x1=cos(squeeze(xall(1,:,iwsort)));
    	y1=sin(squeeze(xall(1,:,iwsort)));
    	x2=x1+2*cos(squeeze(xall(2,:,iwsort)));
    	y2=y1+2*sin(squeeze(xall(2,:,iwsort)));
	case 'singlewell'
    	uex=zeros(1,N);
    	uex(1)=-a/(R+a*T)*x0;
    	uex(2:N)=-a./(R+a*(T-(1:N-1)*dt)).*x(1,1:N-1);
		mean_u(i)=mean(u-uex);
    	std_u(i)=std(u-uex);
	end;
	if (i==1)|(i==n_ref)
		plotroutine;
	end

	nu=noise_factor*nu;
	lambda=R*nu;
nu
	plot_summary
end;
