% general program for learning optimal control (PIRL) as described in section 20.2
clear all
PLANT='singlewell';
PLANT='acrobot';
PLANT='neuralnetwork';
PLANT='lorenz';
NEIGHBORHOOD='size';		% number: set by number of points, size: set by size
dt=0.01;
T=1;						% horizon time
Tlarge=10;
Tinit=10;
Twindow=10;
rho=1;					% initial neighborhood size when NEIGHBORHOOD='size';
rhomax=1;					% maximal neighborhood size when NEIGHBORHOOD='size';

rand('state',2);
randn('state',2);


W=0;
switch PLANT
case 'singlewell'
	n=1;
	m=1;
	x0=2*ones(n,1);
	nu=1;						% diagonal covariance matrix <xi_i xi_j> = nu dt delta_{ij}
	R=1;						% diagonal u^2 cost R sum_i u_i^2
	n_neighbor=100;				% number of neighboring points when NEIGHBORHOOD='number';
	a=5;						% size of quadratic end of path cost
case 'acrobot',
	n=4;					% dimension x
	m=1;					% dimension u, xi
	l1=1;
	l2=2;
	x0=[-pi/2,-pi/2,0,0]';
	nu=5;						% diagonal covariance matrix <xi_i xi_j> = nu dt delta_{ij}
	nufinal=nu;
	
	R=1;						% diagonal u^2 cost R sum_i u_i^2
	n_neighbor=200;				% number of neighboring points when NEIGHBORHOOD='number';
	a=0.5;						% size of quadratic end of path cost relative to control cost
case 'neuralnetwork',
	n=10;
	m=n;
	Wstrength=3;
	W=Wstrength/sqrt(n)*tril(randn(n,n));
	Wanti=0.5*(W-W');
	W=Wstrength/sqrt(n)*tril(rand(n,n));
	Wattr=-0*(W+W');
	W=Wanti+Wattr;
	nu=0.0;
	nufinal=nu;
	R=1;
	x0=1*ones(n,1);
	a=0;
case 'lorenz',
	n=3;
	m=n;
	nu=20;
	nufinal=nu;
	R=1;
	x0=1*ones(n,1);
	a=0;
end;

% end definitions -------------------------------------------------------
N=T/dt;
Nlarge=Tlarge/dt;
Ninit=Tinit/dt;
Nwindow=Twindow/dt;
if nu==0,
	noise_factor=1;
else
	noise_factor=(nufinal/nu)^(1/Nlarge);
end;
lambda=nu*R;

x=zeros(n,Nlarge);			% state history
u=zeros(m,Nlarge);			% control history
xi=zeros(m,Nlarge);			% noise history
dsk=zeros(1,Nlarge);		% instantaneous kinetic path cost of each state
height=zeros(1,Nlarge);		% instantaneous height of each state
velocity=zeros(1,Nlarge);		% instantaneous squared velocity
dsv=zeros(1,Nlarge);		% instantaneous potential path cost of each state
s=zeros(1,Nlarge);			% future path cost of each state
dx1=zeros(1,Nlarge);		% size of neighborhood for each time
l=zeros(1,Nlarge);			% number of states in neighborhood
J=zeros(1,Nlarge);			% expected cost to go
ss=zeros(1,Nlarge);			% effective system size

% initialize 
x(:,1)=x0;

for t=1:Nlarge
	xi(:,t)=sqrt(nu*dt)*randn(m,1);	
	
	% compute u^t
	if t<Ninit,
		u(:,t)=0;
	else
		% compute set of points near x
		tw=max(t-N-Nwindow,1):t-N-1;
		lw=length(tw);
		distance=zeros(m,lw);
		switch PLANT
		case 'acrobot'
			klad=exp(i*x(1:2,t)*ones(1,lw))-exp(i*x(1:2,tw));
			distance(1:2,:)=abs(klad.*conj(klad));
			distance(3:4,:)=abs(x(3:4,t)*ones(1,lw)-x(3:4,tw));
		otherwise
			distance=sum((x(:,t)*ones(1,lw)-x(:,tw)).^2,1);
		end
		maxdistance=max(distance,[],2);
			
		switch NEIGHBORHOOD
		case 'size'
			iset=tw(find(all(distance<=rho*maxdistance*ones(1,lw),1)));
			l(t)=length(iset);
			mins=min(s(iset));
   			w_iset=exp(-(s(iset)-mins)/lambda);
			dx1(t)=rho;
   			while (l(t)==0)&(dx1<rhomax)
   				dx1(t)=dx1(t)+rho;
   				iset=tw(find(all(distance<=dx1(t)*maxdistance*ones(1,lw),1)));
   				l(t)=length(iset);
				mins=min(s(iset));
   				w_iset=exp(-(s(iset)-mins)/lambda);
   			end;
			if (l(t)==0), fprintf('ERROR: no neighbors found\n'); 
				%distance
				%maxdistance
				%iset
				%dx1(t)
				%x
				pause; end;
		case 'number'
			[distance_sort, idistance]=sort(distance, 'ascend');
   			iset=idistance(1:n_neighbor);
			l(t)=length(iset);
			mins=min(s(iset));
   			w_iset=exp(-(s(iset)-mins)/lambda);
			dx1(t)=distance(idistance(n_neighbor));	% largest distance in neighborhood
		end;
		J(t)=mins-lambda*log(sum(w_iset)/l(t));
   		alpha=w_iset/sum(w_iset);
		ss(t)=1/sum(alpha.^2);
		
		x_iset=x(:,iset);			% n times length(iset)
		xi_iset=u(:,iset)*dt+xi(:,iset);		% m times length(iset)
			
		barx=x_iset*alpha';
		barxi=xi_iset*alpha';
		klad=x_iset-barx*ones(1,l(t));
		chi=klad*spdiags(alpha',0,l(t),l(t))*klad';
		deltax=x(:,t)-barx;
		if (l(t)>=size(x,1))&(sum(deltax.^2)>0.01)&&(min(eig(chi))>sum(deltax.^2))
			disp('chi used')
			klad1=xi_iset-barxi*ones(1,length(iset));
   			etaxi=klad1*spdiags(alpha',0,l(t),l(t))*klad';
   			wxi=etaxi*inv(chi);
   			u(:,t)=(wxi*deltax+barxi)/dt;
			[wxi*deltax, barxi]
		else
			u(:,t)=barxi/dt;
		end;
	end;
	if t>1
		%u(:,t)=u(:,t-1)+0.001*u(:,t);
	end;

	% update s of recent states
	dsk(t)=dSk(u(:,t),xi(:,t),R,dt);
	dsv(t)=dSv(x(:,t),a,dt,PLANT);
	switch PLANT,
	case 'acrobot'
		height(t)=l1*sin(x(1,t))+l2*sin(x(2,t));
		velocity(t)=x(3,t)^2+x(4,t)^2;
	end;
	%s(max(1,t-N):t-1)=s(max(1,t-N):t-1)+dsk(t)+dsv(t);
	s(max(1,t-N):t-1)=s(max(1,t-N):t-1)+dsv(t);

	x(:,t+1)=plant(x(:,t),u(:,t)*dt+xi(:,t),dt,PLANT,W);

	nu=noise_factor*nu;
	lambda=R*nu;
end;
figure(20)
t=(1:Nlarge)*dt;
clf
subplot(2,5,1)
plot(t,x(:,1:Nlarge)')
ylabel('x')
subplot(2,5,2)
plot(t,u*dt,t,ones(1,Nlarge)*sqrt(nu*dt),'b-.')
ylabel('u dt')
subplot(2,5,3)
plot(t,l)
ylabel('l')
subplot(2,5,4)
plot(t,dx1)
ylabel('dx1')
subplot(2,5,5)
plot(t,ss)
ylabel('ss')
subplot(2,5,6)
plot(t,J)
ylabel('J')
subplot(2,5,7)
plot(t,dsk/dt)
ylabel('dsk')
subplot(2,5,8)
plot(t,dsv/dt)
ylabel('dsv')
switch PLANT
case 'acrobot'
	subplot(2,5,9)
	plot(t,height)
	ylabel('height')
	subplot(2,5,10)
	plot(t,velocity)
	ylabel('velocity')
	figure(1)
	clf
	times=0.9*Nlarge:Nlarge;
	times=1:500;
	plot(cos(x(1,times))+2*cos(x(2,times)),sin(x(1,times))+2*sin(x(2,times)))
case 'neuralnetwork'
	subplot(2,5,9)
	plot(t,x(1,1:Nlarge))
end;
