% optimal control computation in quadratic potential 
% cost is int_0^T 0.5 R u^2 + 0.5 Q x^2
% dynamics is dx=d\xi
% detailed study of effect of number of samples and nearby data points on estimate
% of u
clear all


R=1;
Q=5;
T=1;		% horizon time
dt=0.01;
N=T/dt;		% number of time points


rand('state',1);
randn('state',1);
x0=2;		% initial point
dx0=0;		% std in initial point
P=sqrt(Q*R)*tanh(sqrt(Q/R)*T);
uex=-P/R*x0;

n=100;
maxiter=100;
maxj=10;
weighted_u=1;
importance=1;	% using importance sampling with optimal u

ss=zeros(maxj,maxiter);
u=zeros(maxj,maxiter);
J=zeros(maxj,maxiter);
psi=zeros(maxj,maxiter);
meanu=zeros(1,maxj);
stdu=zeros(1,maxj);
meanss=zeros(1,maxj);
stdss=zeros(1,maxj);
meanJ=zeros(1,maxj);
stdJ=zeros(1,maxj);



	
for j=1:maxj,
	nu(j)=0.2*j;
	lambda(j)=R*nu(j);
	sigma=sqrt(nu(j)*dt);
	beta=0.5*lambda(j)*log(cosh(sqrt(Q/R)*T));
	Jex(j)=0.5*P*x0^2+beta;
	psiex(j)=exp(-Jex(j)/lambda(j));
	for i=1:maxiter,
		u0=zeros(n,N);
		xiall=sigma*randn(n,N);
		xall=zeros(n,N);
		x0all=x0+dx0*randn(n,1);		% initial point of trajectory
		%if importance, u0(:,1)=-P/R*x0all; end;
		if importance, u0(:,1)=-1; end;
		xall(:,1)=x0all+u0(:,1)*dt+xiall(:,1);
		for t=2:N,
			%if importance, u0(:,t)=-P/R*xall(:,t-1); end;
			if importance, u0(:,t)=-1; end;
			xall(:,t)=xall(:,t-1)+u0(:,t)*dt+xiall(:,t);
		end;
		Sv=0.5*dt*Q*sum(xall(:,1:N).^2,2);
		Sk=0.5*R*dt*sum(u0.^2,2)+R*sum(xiall.*u0,2);
		%Sk=0.5*R*dt*sum(u0.^2,2);
		S=Sv+Sk;	
		w=exp(-S/lambda(j));

		% compute psi 
		barx=mean(x0all);
		chi=1/n*sum((x0all-barx).^2);
		if dx0>0
			barS=sum(S)/n;
			etaS=1/n*S'*(x0all-barx);
			wS=inv(chi)*etaS;
			psi(j,i)=1/n*sum(exp(-(S+wS*(x0-x0all))/lambda(j)));
		else
			psi(j,i)=sum(w)/n;
		end;
		J(j,i)=-lambda(j)*log(psi(j,i));

		% compute <dxi>
		alpha=w/sum(w);
		barxi=alpha'*xiall(:,1);
		if dx0>0,
			barx=alpha'*x0all;
			chi=alpha'*((x0all(:,1)-barx).^2);
			eta=alpha'*((xiall(:,1)-barxi).*(x0all(:,1)-barx));
			wxi=eta*inv(chi);
			dxi=wxi*(x0-barx)+barxi;

			% local model for u0
			if weighted_u,
        		baru=alpha'*u0(:,1);
        		etau=alpha'*((u0(:,1)-baru).*(x0all(:,1)-barx));
        		wu=etau*inv(chi);
			else
        		baru=1/n*sum(u0(:,1));
        		etau=1/n*sum((u0(:,1)-baru).*(x0all(:,1)-barx));
        		wu=etau*inv(chi);
			end;
        	u0x=wu*(x0-barx)+baru;
		else
			dxi=barxi;
			if weighted_u,
				u0x=alpha'*u0(:,1);
			else
				u0x=1/n*sum(u0(:,1));
			end;
		end;
		u(j,i)=dxi/dt+u0x;

		% effective sample size
		ss(j,i)=1/sum(alpha.^2);
	end;
end;
meanu=mean(u,2);
stdu=std(u,[],2);
meanss=mean(ss,2);
stdss=std(ss,[],2);
meanJ=mean(J,2);
stdJ=std(J,[],2);
meanpsi=mean(psi,2);
stdpsi=std(psi,[],2);
%figure(2)
%hist(psi(1,:),100)
figure(1)
clf
subplot(2,4,1)
errorbar(nu,meanss,stdss)
xlabel('\nu')     
ylabel('sample size')
%axis([0 2.5 0 40])
subplot(2,4,2)
errorbar(nu,meanu,stdu)
hold on
plot(nu,ones(1,maxj)*uex,'r')
xlabel('\nu')
ylabel('u')
%axis([0 2.5 -10 0])
hold off
subplot(2,4,3)
errorbar(nu,meanJ,stdJ)
hold on
plot(nu,Jex,'r');
xlabel('\nu')
ylabel('J')
%axis([0 2.5 4 6.5])
hold off
subplot(2,4,4)
errorbar(nu,-lambda'.*log(meanpsi),lambda'.*stdpsi./meanpsi)
hold on
plot(nu,-lambda.*log(psiex),'r');
xlabel('\nu')
ylabel('-\lambda log \psi ')
%axis([0 2.5 4 6.5])
hold off
-lambda'.*log(meanpsi)+lambda'.*log(psiex)'
